#!/usr/bin/python
# -encoding:utf8-

# @file origami-visualize.py
# Executable visualising an origami, compatible with visualize.cpp.
# 
# @author Joachim Breitner
# @version $Id: origami-visualize.py 427 2011-06-08 21:12:39Z joachim $
# @date 29.4.2009

#  This file is part of the Origami-Library, which is a collection
#  of algorithms that calculate properties of mathematical objects
#  called "origamis". It is developed within the project "With origamis to
#  Teichmueller curves in moduli space".
#
#  Copyright (C) 2009,2011 Joachim Breitner
#  Address: Institute for Algebra and Geometry, Fakultaet fuer
#  Mathematik, Universitaet Karlsruhe, D-76128 Karlsruhe, Germany.
#  E-mail: mail@joachim-breitner.de or origami@math.uni-karlsruhe.de
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 3
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with this program; if not, see
#  <http://www.gnu.org/licenses/> or write to the Free
#  Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA  02110-1301, USA.
#
#  If you publish results you obtained by using this software please
#  don't forget to give appropriate credit in your publication.

# Import statements
import re
import pygtk
pygtk.require('2.0')
import gtk
import gobject
import cairo
import sys
import os.path
from math import floor, ceil

# Graphics constants
screenSize = 20 # in squares
squareSize = 100

# Some global constants
RIGHT=0
UP=1
LEFT=2
DOWN=3

reverse = { RIGHT: LEFT, LEFT: RIGHT, UP: DOWN, DOWN: UP}

dirs = (RIGHT, UP, LEFT, DOWN)

# General helper functions
def neighbours_of_pos((x,y)):
    # The order of directions is the same as in visualization.cpp
    return  [(RIGHT,(x+1,y)),
             (UP,   (x,y+1)),
             (LEFT, (x-1,y)),
             (DOWN, (x,y-1)),
             ]

class Graph:
    def __init__(self,size):
        self.size = size
        self.neigh = [ {},{},{},{} ]

    @classmethod
    def load(cls, filename):
        f = file(filename)
        
        graph = cls(int(f.readline()))
        
        for i in range(0,graph.size):
            (name, right, top) = map(int, f.readline().split())
            graph.neigh[RIGHT][name] = right
            graph.neigh[LEFT][right] = name
            graph.neigh[UP][name] = top
            graph.neigh[DOWN][top] = name

        return graph


class Visualisation:
    def __init__(self,graph):
        self.graph = graph
        self.pos = {}

    def is_complete(self):
        for name in self.graph.neigh[RIGHT].keys():
            if name not in self.pos:
                return False
        return True

    def is_free(self,pos):
        return pos not in self.pos.values()

    def vertex_at(self, pos):
        for v, p in self.pos.iteritems():
            if p == pos:
                return v

    def candidates(self, pos):
        candidates = []
        for (dir, other_pos) in neighbours_of_pos(pos):
            if not self.is_free(other_pos):
                other = self.vertex_at(other_pos)
                other_neigh = self.graph.neigh[reverse[dir]][other]
                if other_neigh not in candidates:
                    candidates.append(other_neigh)
        return candidates

    def is_fillable(self, pos):
        return len(self.candidates(pos)) == 1

    def conflict_at(self, pos):
        return len(self.candidates(pos)) > 1

    def fill(self,pos):
        assert self.is_fillable(pos)
        self.pos[self.candidates(pos)[0]] = pos

    def remove_at(self,pos):
        name = self.vertex_at(pos)
        del self.pos[name]
        return name

    def save(self, filename):
        f = file(filename, 'w')

        # Ensure that all saved coordinates are non-negative
        min_x, min_y = 0, 0
        for name, (x,y) in self.pos.iteritems():
            min_x = min(min_x,x)
            min_y = min(min_y,y)
        

        f.write('%% Origami of size %d\n' % self.graph.size)
        f.write('\\begin{xy}\n')
        f.write('<0.8cm,0cm>:\n')
        for name, (x,y) in self.pos.iteritems():
            f.write('(%d,%d)*{\\OriSquare{%s}' % (x-min_x,y-min_y,name))
            for (dir,other_pos) in neighbours_of_pos((x,y)):
                if self.is_free(other_pos):
                    f.write('{%s}' % self.graph.neigh[dir][name])
                else:
                    f.write('{}')
            f.write('};\n')
        f.write('\\end{xy}\n')

    @classmethod
    def load(cls, graph, filename):
        f = file(filename)
        
        size = int(re.match('% Origami of size (\d+)',f.readline()).group(1))
        f.readline() # \begin{xy}
        f.readline() # <0.8cm,0cm>:

        data = []
        for _ in range(0,size):
            data.append(
                   re.match(
                        r'\(([-\d]+),([-\d]+)\)\*{\\OriSquare{(\d+)}{(\d*)}{(\d*)}{(\d*)}{(\d*)}};',
                        f.readline()
                   ).groups()
                );
        
        #graph = Graph(size)
        #for (_,_, name, right, top, _, _) in data
        #    graph.neigh[RIGHT][name] = right
        #    graph.neigh[LEFT][right] = name
        #    graph.neigh[UP][name] = top
        #    graph.neigh[DOWN][top] = name

        visual = cls(graph)
        for (x,y, name,_ ,_ ,_ ,_ ) in data:
            assert int(name) in graph.neigh[RIGHT]
            visual.pos[int(name)] = (int(x),int(y))

        return visual

class GtkVisualize:
    def __init__(self):
        self.zoom = 1.0

        # GUI-Stuff
        ag = gtk.AccelGroup()

        self.draw = gtk.Layout()
        self.draw.set_size(int(ceil(self.zoom * 2 * screenSize * squareSize)),
                           int(ceil(self.zoom * 2 * screenSize * squareSize)))
        self.draw.add_events(gtk.gdk.BUTTON_PRESS_MASK |
                        gtk.gdk.BUTTON_RELEASE_MASK |
                        gtk.gdk.LEAVE_NOTIFY_MASK |
                        gtk.gdk.EXPOSURE_MASK |
                        gtk.gdk.POINTER_MOTION_MASK |
                        gtk.gdk.KEY_PRESS_MASK )
        self.draw.connect("button-press-event",self.click)
        self.draw.connect("button-release-event",self.unclick)
        self.draw.connect("motion-notify-event",self.hover)
        self.draw.connect_after("expose-event",self.do_expose_event)
        self.draw.connect("key-press-event", self.key_press)
        self.draw.props.can_focus = True
     
        self.scrolledwin = gtk.ScrolledWindow()
        self.scrolledwin.get_hadjustment().props.step_increment = 0.2 * squareSize
        self.scrolledwin.get_vadjustment().props.step_increment = 0.2 * squareSize
        self.scrolledwin.add(self.draw)

        self.vertex_list = gtk.ListStore(gobject.TYPE_INT, gobject.TYPE_BOOLEAN)
        self.tv = gtk.TreeView(self.vertex_list)
        self.tv.set_size_request(100,-1)
        column = gtk.TreeViewColumn("Vertex")
        cr = gtk.CellRendererText()
        column.pack_start(cr)
        def append_checkmark(column, cell, model, iter):
            cell.props.text = model.get_value(iter, 0)
            if model.get_value(iter, 1):
                cell.props.text += " ✓"
        column.set_cell_data_func(cr, append_checkmark)
        self.tv.append_column(column)
        self.tv.get_selection().connect("changed", self.scroll_to_selection)

        sw2 = gtk.ScrolledWindow()
        sw2.set_policy(gtk.POLICY_NEVER,gtk.POLICY_AUTOMATIC)
        sw2.add(self.tv)

        hbox = gtk.HBox()
        hbox.pack_start(self.scrolledwin)
        hbox.pack_start(sw2, expand=False)

        openGraphButton = gtk.ToolButton(label = "Open .ori")
        openGraphButton.set_stock_id(gtk.STOCK_OPEN)
        openGraphButton.set_is_important(True)
        openGraphButton.connect("clicked",self.openGraph)
        openGraphButton.add_accelerator("clicked", ag, ord('o'), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        openGraphButton.set_tooltip_text('Open Origmai (Ctrl-O)')

        newVisButton = gtk.ToolButton(label = "New .vis")
        newVisButton.set_stock_id(gtk.STOCK_NEW)
        newVisButton.connect("clicked",self.newVis)
        newVisButton.add_accelerator("clicked", ag, ord('n'), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        newVisButton.set_tooltip_text('New visualization (Ctrl-N)')

        openVisButton = gtk.ToolButton(label = "Open .vis")
        openVisButton.set_stock_id(gtk.STOCK_OPEN)
        openVisButton.set_is_important(True)
        openVisButton.connect("clicked",self.openVis)
        openVisButton.add_accelerator("clicked", ag, ord('v'), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        openVisButton.set_tooltip_text('Open visualization (Ctrl-V)')

        saveVisButton = gtk.ToolButton(label = "Save .vis")
        saveVisButton.set_stock_id(gtk.STOCK_SAVE)
        saveVisButton.connect("clicked",self.saveVis)
        saveVisButton.add_accelerator("clicked", ag, ord('s'), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        saveVisButton.set_tooltip_text('Save visualization (Ctrl-S)')

        saveVisAsButton = gtk.ToolButton(label = "Save .vis as..")
        saveVisAsButton.set_stock_id(gtk.STOCK_SAVE_AS)
        saveVisAsButton.connect("clicked",self.saveVisAs)
        saveVisAsButton.add_accelerator("clicked", ag, ord('s'), gtk.gdk.CONTROL_MASK |  gtk.gdk.SHIFT_MASK, gtk.ACCEL_VISIBLE)
        saveVisAsButton.set_tooltip_text('Save visualization as... (Ctrl-Shift-V)')

        zoomInButton = gtk.ToolButton(label = "Zoom in")
        zoomInButton.set_stock_id(gtk.STOCK_ZOOM_IN)
        zoomInButton.connect("clicked",self.changeZoom,1)
        zoomInButton.add_accelerator("clicked", ag, ord('+'), 0, gtk.ACCEL_VISIBLE)
        zoomInButton.set_tooltip_text('Zoom in (+)')
        zoomOutButton = gtk.ToolButton(label = "Zoom out")
        zoomOutButton.set_stock_id(gtk.STOCK_ZOOM_OUT)
        zoomOutButton.connect("clicked",self.changeZoom,-1)
        zoomOutButton.add_accelerator("clicked", ag, ord('-'), 0, gtk.ACCEL_VISIBLE)
        zoomOutButton.set_tooltip_text('Zoom out (-)')
        zoomFitButton = gtk.ToolButton(label = "Zoom to fit")
        zoomFitButton.set_stock_id(gtk.STOCK_ZOOM_FIT)
        zoomFitButton.connect("clicked",self.zoomToFit)
        zoomFitButton.set_tooltip_text('Zoom to fit visualization')


        quitButton = gtk.ToolButton(gtk.STOCK_QUIT)
        quitButton.add_accelerator("clicked", ag, ord('q'), gtk.gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
        quitButton.set_tooltip_text('Quit (Ctrl-q)')

        toolbar = gtk.Toolbar()
        toolbar.set_style(gtk.TOOLBAR_BOTH_HORIZ)
        toolbar.add(openGraphButton)
        toolbar.add(gtk.SeparatorToolItem())
        toolbar.add(newVisButton)
        toolbar.add(openVisButton)
        toolbar.add(saveVisButton)
        toolbar.add(saveVisAsButton)
        toolbar.add(gtk.SeparatorToolItem())
        toolbar.add(zoomInButton)
        toolbar.add(zoomOutButton)
        toolbar.add(zoomFitButton)
        toolbar.add(gtk.SeparatorToolItem())
        toolbar.add(quitButton)

        vbox = gtk.VBox()
        vbox.pack_start(toolbar, expand=False)
        vbox.pack_start(hbox)

        self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        self.window.set_default_size(900,700)
        self.window.set_title("Origami visualizer")
        self.window.connect("delete_event", self.delete_event)
        self.window.connect("destroy", self.destroy)
        self.window.connect("size-allocate", self.first_scroll)
        self.window.add_accel_group(ag)
        self.starting = True
        def done_starting():
            self.starting = False
            return False
        gobject.timeout_add(500, done_starting)
        self.window.add(vbox)

        quitButton.connect("clicked",lambda _: self.window.destroy())

        self.pan_last_event = None

        # Non GUI-Stuff
        self.graph = None
        self.graph_filename = None
        self.visual = None
        self.visual_filename = None

        self.window.show_all()
        
    def scroll_to_selection(self, widget):
        name = self.getSelectedVertex()
        if name is not None:
            self.scroll_to_vertex(name)

    def first_scroll(self, widget, event):
        # crude hack
        if self.starting:
            self.scroll_to_center()

    def scroll_to_center(self):
        max_x = -screenSize * squareSize
        min_x =  screenSize * squareSize
        max_y = -screenSize * squareSize
        min_y =  screenSize * squareSize
        if self.visual:
            for pos in self.visual.pos.values():
                (x,y) = self.posToCenter(pos)
                max_x = max(max_x,x)
                min_x = min(min_x,x)
                max_y = max(max_y,y)
                min_y = min(min_y,y)
        x =(max_x + min_x)/2
        y =(max_y + min_y)/2
        
        x += screenSize * squareSize
        y += screenSize * squareSize

        x *= self.zoom
        y *= self.zoom

        size = self.scrolledwin.get_allocation()
        self.scrolledwin.get_hadjustment().props.value = \
            x - self.scrolledwin.get_hadjustment().props.page_size/2
        self.scrolledwin.get_vadjustment().props.value = \
            y - self.scrolledwin.get_vadjustment().props.page_size/2

    def scroll_to_vertex(self, name):
        if name in self.visual.pos:
            (x,y) = self.posToCenter(self.visual.pos[name])
            x += screenSize * squareSize
            y += screenSize * squareSize

            x *= self.zoom
            y *= self.zoom

            # check if it is visible already
            size = self.scrolledwin.get_allocation()
            offset_x = self.scrolledwin.get_hadjustment().props.value
            offset_y = self.scrolledwin.get_vadjustment().props.value
            if not offset_x <= x <= offset_x + size.width:
                self.scrolledwin.get_hadjustment().props.value = x - size.width/2
            if not offset_y <= y <= offset_y + size.height:
                self.scrolledwin.get_vadjustment().props.value = y - size.height/2
        self.draw.queue_draw()

    def openGraph(self, widget=None):
        dialog = gtk.FileChooserDialog("Open origami...", None, gtk.FILE_CHOOSER_ACTION_OPEN,
                                      (gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,
                                       gtk.STOCK_OPEN, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)
        filter = gtk.FileFilter()
        filter.set_name("Origami")
        filter.add_pattern("*.ori")
        dialog.add_filter(filter)
        if self.graph_filename:
            dialog.set_filename(self.graph_filename)
        elif os.path.exists("examples/origamis"):
            dialog.set_filename("examples/origamis/*.ori")

        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            self.loadGraph(dialog.get_filename())
        dialog.destroy()

    def openVis(self, widget=None):
        if self.graph:
            dialog = gtk.FileChooserDialog("Open origami visualization...",
                                          None, gtk.FILE_CHOOSER_ACTION_OPEN,
                                          (gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,
                                           gtk.STOCK_OPEN,  gtk.RESPONSE_OK))
            dialog.set_default_response(gtk.RESPONSE_OK)
            filter = gtk.FileFilter()
            filter.set_name("Origami visualization")
            filter.add_pattern("*.vis")
            dialog.add_filter(filter)
            if self.visual_filename:
                dialog.set_filename(self.visual_filename)
            elif self.graph_filename:
                dialog.set_filename(self.graph_filename.replace(".ori",".vis"))

            response = dialog.run()
            if response == gtk.RESPONSE_OK:
                self.loadVisualisation(dialog.get_filename())
            dialog.destroy()

    def checkCompleteness(self):
        if not self.visual.is_complete():
            dialog = gtk.MessageDialog(
                    parent = self.window,
                    flags = gtk.DIALOG_MODAL,
                    type=gtk.MESSAGE_WARNING,
                    buttons=gtk.BUTTONS_OK,
                    message_format="The visualization does not include all "+
                        "vertices of the graph.")
            dialog.run()
            dialog.destroy()

    def saveVis(self, widget=None):
        if self.visual:
            if self.visual_filename:
                self.checkCompleteness()
                self.visual.save(self.visual_filename)
            else:
                self.saveVisAs()

    def saveVisAs(self, widget=None):
        if self.graph:
            self.checkCompleteness()
            dialog = gtk.FileChooserDialog("Save origami visualization...",
                                          None, gtk.FILE_CHOOSER_ACTION_SAVE,
                                          (gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,
                                           gtk.STOCK_SAVE, gtk.RESPONSE_OK))
            dialog.set_default_response(gtk.RESPONSE_OK)
            filter = gtk.FileFilter()
            filter.set_name("Origami visualization")
            filter.add_pattern("*.vis")
            dialog.add_filter(filter)
            if self.visual_filename:
                dialog.set_filename(self.visual_filename)
            elif self.graph_filename:
                suggested = self.graph_filename.replace(".ori",".vis")
                dialog.set_current_folder(os.path.dirname(suggested))
                dialog.set_current_name(os.path.basename(suggested))

            response = dialog.run()
            if response == gtk.RESPONSE_OK:
                self.visual_filename = dialog.get_filename()
                self.visual.save(self.visual_filename)
            dialog.destroy()

    def newVis(self, widget=None):
        if self.graph:
            self.visual_filename = None
            self.visual = Visualisation(self.graph)
            self.vertex_list.clear()
            for name in self.graph.neigh[RIGHT].keys():
                self.vertex_list.append((name,False))
            self.draw.queue_draw()

            if self.getSelectedVertex() is None:
                self.tv.get_selection().select_path(0)


    def loadGraph(self, filename):
        self.graph_filename = filename
        self.graph = Graph.load(filename)
        self.vertex_list.clear()
        for name in self.graph.neigh[RIGHT].keys():
            self.vertex_list.append((name,False))
        self.newVis()
        self.draw.queue_draw()

    def loadVisualisation(self, filename):
        self.visual_filename = filename
        self.visual = Visualisation.load(self.graph, filename)
        for name in self.visual.pos:
            self.setVertexListCheck(name, True)
        self.scroll_to_center()

    def zoomToFit(self, widget):
        max_x = -screenSize
        min_x =  screenSize
        max_y = -screenSize
        min_y =  screenSize
        if self.visual:
            for x,y in self.visual.pos.values():
                max_x = max(max_x,x)
                min_x = min(min_x,x)
                max_y = max(max_y,y)
                min_y = min(min_y,y)
        width = max_x - min_x + 2
        height = max_y - min_y + 2

        if width > 0 and height > 0:
            self.zoom = min(float(self.scrolledwin.get_hadjustment().props.page_size) / (width * squareSize),
                            float(self.scrolledwin.get_vadjustment().props.page_size) / (height * squareSize))
            self.draw.set_size(int(ceil(self.zoom * 2 * screenSize * squareSize)),
                               int(ceil(self.zoom * 2 * screenSize * squareSize)))
            self.scroll_to_center()
            self.draw.queue_draw()
        

    def changeZoom(self, widget, power):
        base = 1.25
        self.zoom *= base**power
        self.draw.set_size(int(ceil(self.zoom * 2 * screenSize * squareSize)),
                           int(ceil(self.zoom * 2 * screenSize * squareSize)))
        size = self.scrolledwin.get_allocation()
        self.scrolledwin.get_hadjustment().props.value = \
            (self.scrolledwin.get_hadjustment().props.value +
             self.scrolledwin.get_hadjustment().props.page_size/2) * base**power \
            - self.scrolledwin.get_hadjustment().props.page_size/2
        self.scrolledwin.get_vadjustment().props.value = \
            (self.scrolledwin.get_vadjustment().props.value +
             self.scrolledwin.get_vadjustment().props.page_size/2) * base**power \
            - self.scrolledwin.get_vadjustment().props.page_size/2
        self.draw.queue_draw()

    def setVertexListCheck(self, name, val):
        for row in self.vertex_list:
            if row[0] == name:
                row[1] = val

    def getSelectedVertex(self):
        sel = self.tv.get_selection()
        (model,iter) = sel.get_selected()
        if iter is not None:
            return model.get_value(iter, 0)

    def isVertexSelcted(self, name):
        return name == self.getSelectedVertex()

    def key_press(self, draw, event):
        if not self.graph:
            return

        if event.state == 0:
            keyname = gtk.gdk.keyval_name(event.keyval)
            if keyname in ("Left", "Right", "Up", "Down"):
                sel = self.getSelectedVertex()
                if sel is not None:
                    (x,y) = self.visual.pos[sel]
                    if   keyname == "Left":
                        other_pos = (x-1,y)
                    elif keyname == "Right":
                        other_pos = (x+1,y)
                    elif keyname == "Up":
                        other_pos = (x,y+1)
                    elif keyname == "Down":
                        other_pos = (x,y-1)
                    self.activate_square(other_pos)
                return True

    def activate_square(self,pos):
                if self.visual.is_free(pos):
                    if self.visual.conflict_at(pos):
                        pass
                    elif self.visual.is_fillable(pos):
                        self.visual.fill(pos)
                        self.setVertexListCheck(self.visual.vertex_at(pos), True)
                        self.draw.queue_draw()
                    else:
                        name = self.getSelectedVertex()
                        if name is not None:
                            self.visual.pos[name] = pos
                            self.setVertexListCheck(name, True)
                            self.draw.queue_draw()

                for row in self.vertex_list:
                    if row[0] in self.visual.pos and self.visual.pos[row[0]] == pos:
                        self.tv.get_selection().select_iter(row.iter)

    def click(self, draw, event):
        draw.grab_focus()

        if not self.graph:
            return

        pos = self.clickToPos((event.x, event.y))

        if event.button == 1:
            if event.get_state() & gtk.gdk.CONTROL_MASK:
                if not self.visual.is_free(pos):
                    name = self.visual.remove_at(pos)
                    self.setVertexListCheck(name, False)
                    self.draw.queue_draw()
            else:
                self.activate_square(pos)
        elif event.button == 2:
            draw.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.FLEUR))
            self.pan_last_event = \
                (event.get_root_coords()[0] + self.scrolledwin.get_hadjustment().props.value,
                 event.get_root_coords()[1] + self.scrolledwin.get_vadjustment().props.value)

    def unclick(self,widget,event):
        self.pan_last_event = None
        widget.window.set_cursor(None)

    def hover(self,widget,event):
        if self.pan_last_event is not None and event.state & gtk.gdk.BUTTON2_MASK:
            self.scrolledwin.get_hadjustment().props.value = \
                -event.get_root_coords()[0] + self.pan_last_event[0]
            self.scrolledwin.get_vadjustment().props.value = \
                -event.get_root_coords()[1] + self.pan_last_event[1]


    def clickToPos(self, (x,y)):
        return ( int(floor( (x/self.zoom - screenSize * squareSize  - squareSize/2) / squareSize)),
                 int(floor(-(y/self.zoom - screenSize * squareSize  + squareSize/2) / squareSize)))

    def posToCorner(self, (x,y)):
        return ( x * squareSize  + squareSize/2,
                 -y * squareSize - squareSize/2)

    def posToCenter(self, (x,y)):
        return ( x  * squareSize  + 2*squareSize/2,
                 -y * squareSize - 2*squareSize/2)

    def triangles_from_pos(self, (x,y)):
        cx,cy = self.posToCenter((x,y))
        return [ (UP,   (x,y+1),
                    [(cx - squareSize/2 + 0.1 *squareSize, cy -     squareSize/2 - 0.05*squareSize),
                     (cx                                 , cy -   2*squareSize/2 + 0.05*squareSize),
                     (cx + squareSize/2 - 0.1 *squareSize, cy -     squareSize/2 - 0.05*squareSize)],
                     (cx                                 , cy - 1.4*squareSize/2)),
                 (RIGHT,(x+1,y),
                    [(cx + squareSize/2 + 0.05*squareSize, cy -     squareSize/2 + 0.1* squareSize),
                     (cx + squareSize   - 0.05*squareSize, cy),
                     (cx + squareSize/2 + 0.05*squareSize, cy +     squareSize/2 - 0.1* squareSize)],
                     (cx + 1.4*squareSize/2              , cy)),
                 (DOWN, (x,y-1),
                    [(cx - squareSize/2 + 0.1 *squareSize, cy +     squareSize/2 + 0.05*squareSize),
                     (cx                                 , cy +   2*squareSize/2 - 0.05*squareSize),
                     (cx + squareSize/2 - 0.1 *squareSize, cy +     squareSize/2 + 0.05*squareSize)],
                     (cx                                 , cy + 1.4*squareSize/2)),
                 (LEFT, (x-1,y),
                    [(cx - squareSize/2 - 0.05*squareSize, cy -     squareSize/2 + 0.1 *squareSize),
                     (cx - squareSize   + 0.05*squareSize, cy),
                     (cx - squareSize/2 - 0.05*squareSize, cy +     squareSize/2 - 0.1 *squareSize)],
                     (cx - 1.4*squareSize/2              , cy)),
                 ]

    def do_expose_event(self, draw, event):
        gc = draw.bin_window.new_gc()
        gc.set_clip_rectangle(event.area)

        size = draw.get_allocation()

        cr = draw.bin_window.cairo_create()
        # set a clip region for the expose event
        cr.rectangle(event.area.x, event.area.y, event.area.width, event.area.height)
        cr.clip()

        cr.scale(self.zoom, self.zoom)
        cr.translate(screenSize * squareSize, screenSize * squareSize)

        # Grey background
        cr.set_source_rgb(0.8,0.8,0.8)
        cr.paint()

        # Grid
        cr.set_source_rgb(0.2,0.2,0.2)
        cr.set_line_width(0.5)
        for x in range(- screenSize * squareSize, screenSize * squareSize, 23):
            cr.move_to(x + 0.5,-screenSize * squareSize)
            cr.line_to(x + 0.5, screenSize * squareSize)
            cr.stroke()

        for y in range(- screenSize * squareSize, screenSize * squareSize, 23):
            cr.move_to(-screenSize * squareSize,-y - 0.5)
            cr.line_to( screenSize * squareSize,-y - 0.5)
            cr.stroke()

        if self.visual:
            for (name, pos) in self.visual.pos.iteritems():
                cr.rectangle(self.posToCorner(pos)[0]+0.5,
                             self.posToCorner(pos)[1]+0.5,
                             squareSize, -squareSize)

                if self.isVertexSelcted(name):
                    cr.set_source_rgb(0,0,1)
                else:
                    cr.set_source_rgb(1,1,1)
                cr.fill_preserve()
                cr.set_source_rgb(0,0,0)
                cr.set_line_width(0.5)
                cr.stroke()
                cr.set_font_size(0.5*squareSize)
                (_,_,w,h,_,_) = cr.text_extents(str(name))
                cr.move_to(*self.posToCenter(pos))
                cr.rel_move_to(-w/2,h/2)
                cr.show_text(str(name))

                for (dir,neigh_pos,triangle,triangle_center) in self.triangles_from_pos(pos):
                    if self.visual.is_free(neigh_pos):
                        neigh = self.visual.graph.neigh[dir][name]

                        cr.move_to(*triangle[0])
                        cr.line_to(*triangle[1])
                        cr.line_to(*triangle[2])
                        cr.close_path()

                        if self.visual.conflict_at(neigh_pos):
                            cr.set_source_rgb(0.4,0.4,0.4)
                        elif neigh in self.visual.pos:
                            cr.set_source_rgb(0,1,0)
                        else:
                            cr.set_source_rgb(1,0,0)
                        cr.fill_preserve()
                        cr.set_source_rgb(0,0,0)
                        cr.set_line_width(0.5)
                        cr.stroke()
                        cr.set_source_rgb(1,1,1)
                        cr.set_font_size(0.2*squareSize)
                        (_,_,w,h,_,_) = cr.text_extents(str(neigh))
                        cr.move_to(*triangle_center)
                        cr.rel_move_to(-w/2,h/2)
                        cr.show_text(str(neigh))

    def delete_event(self, widget, event, data=None):
        # Change FALSE to TRUE and the main window will not be destroyed
        # with a "delete_event".
        return False

    def destroy(self, widget, data=None):
        gtk.main_quit()


    def main(self):
        gtk.main()



if __name__ == "__main__":
    gv = GtkVisualize()
    if len(sys.argv) > 1:
        gv.loadGraph(sys.argv[1])
        if len(sys.argv) > 2:
            gv.loadVisualisation(sys.argv[2])
    gv.main()

# vim:sw=4:sts=4:et
